/** \file
 * \author Daniel Trebbien
 * \par License
 * Public Domain
 */
#include <cassert>
#include <inttypes.h>
#define std_IntegralType typename

// Declarations ///////////////////////////////////////////////////////////////

/*
int_t res;
atomic {
	res = *pI;
	if(*pI != 0)
		--*pI;
}
return res;
*/
template <std_IntegralType int_t>
int_t atomic_dec_if_nz(int_t *pI) throw();

/*
int_t res;
atomic {
	res = *pI;
	if(*pI != std::numeric_limits<int_t>::max())
		++*pI;
}
return res;
*/
template <std_IntegralType int_t>
int_t atomic_inc_if_nmax(int_t *pI) throw();

/*
int_t res;
atomic {
	res = *pI;
	if(*pI != 0 && *pI != std::numeric_limits<int_t>::max())
		++*pI;
}
return res;
*/
template <std_IntegralType int_t>
int_t atomic_inc_if_nz_nmax(int_t *pI) throw();

// Implementations ////////////////////////////////////////////////////////////

template <>
inline uint8_t atomic_dec_if_nz<uint8_t>(uint8_t *pB) throw()
{
	assert(*pB >= 0);

	uint8_t res, tmp;
	asm("movb %[b], %%al\n\t"
		"0: test %%al, %%al\n\t"
		"jz 1f\n\t"
		"movb %%al, %[tmp]\n\t"
		"decb %[tmp]\n\t"
		"lock cmpxchgb %[tmp], %[b]\n\t"
		"jne 0b\n\t"
		"1: movb %%al, %[res]\n\t"
		: [b]"=m"(*pB), [res]"=&r"(res), [tmp]"=&r"(tmp)
		:
		: "al"
		);
	
	return res;
}

template <>
inline uint8_t atomic_inc_if_nmax<uint8_t>(uint8_t *pB) throw()
{
	uint8_t res, tmp;
	asm("movb %[b], %%al\n\t"
		"0: cmp $255, %%al\n\t"
		"je 1f\n\t"
		"movb %%al, %[tmp]\n\t"
		"incb %[tmp]\n\t"
		"lock cmpxchgb %[tmp], %[b]\n\t"
		"jne 0b\n\t"
		"1: movb %%al, %[res]\n\t"
		: [b]"=m"(*pB), [res]"=&r"(res), [tmp]"=&r"(tmp)
		:
		: "al"
		);
	
	return res;
}

template <>
inline uint8_t atomic_inc_if_nz_nmax<uint8_t>(uint8_t *pB) throw()
{
	uint8_t res, tmp;
	asm("movb %[b], %%al\n\t"
		"0: test %%al, %%al\n\t"
		"jz 1f\n\t"
		"cmp $255, %%al\n\t"
		"je 1f\n\t"
		"movb %%al, %[tmp]\n\t"
		"incb %[tmp]\n\t"
		"lock cmpxchgb %[tmp], %[b]\n\t"
		"jne 0b\n\t"
		"1: movb %%al, %[res]\n\t"
		: [b]"=m"(*pB), [res]"=&r"(res), [tmp]"=&r"(tmp)
		:
		: "al"
		);
	
	return res;
}

// Test ///////////////////////////////////////////////////////////////////////
#include <iostream>
#include <limits>
#include <new>

int main()
{
	uint8_t *pB = new uint8_t(3);
	assert(atomic_dec_if_nz<uint8_t>(pB) == 3); assert(*pB == 2);
	assert(atomic_dec_if_nz<uint8_t>(pB) == 2); assert(*pB == 1);
	assert(atomic_dec_if_nz<uint8_t>(pB) == 1); assert(*pB == 0);
	assert(atomic_dec_if_nz<uint8_t>(pB) == 0); /* afterward `*pB` remains 0 because before it was 0 */ assert(*pB == 0);
	assert(atomic_inc_if_nz_nmax<uint8_t>(pB) == 0); /* afterward `*pB` remains 0 because before it was 0 */ assert(*pB == 0);
	assert(atomic_inc_if_nmax<uint8_t>(pB) == 0); assert(*pB == 1);
	assert(atomic_inc_if_nz_nmax<uint8_t>(pB) == 1); assert(*pB == 2);
	*pB = 254;
	assert(atomic_inc_if_nmax<uint8_t>(pB) == 254); assert(*pB == 255);
	assert(atomic_inc_if_nmax<uint8_t>(pB) == 255); /* afterward `*pB` remains 255 because before it was 255 */ assert(*pB == 255);
	assert(atomic_inc_if_nz_nmax<uint8_t>(pB) == 255); /* afterward `*pB` remains 255 because before it was 255 */ assert(*pB == 255);
	assert(atomic_dec_if_nz<uint8_t>(pB) == 255); assert(*pB == 254);
	
	std::cout << "All tests succeeded." << std::endl;
	assert(false);
}
